package App::find2perl;
$App::find2perl::VERSION = '1.004';
1;

__END__

=head1 NAME

App::find2perl - translate find command lines to Perl code

=head1 SYNOPSIS

See L<find2perl>.

=head1 DESCRIPTION

This is a placeholder distribution for CPAN. See L<find2perl> for all functionality.

=cut
